#include <gl/gl.h>
#include <gl/device.h>
#include <math.h>
#include <stdio.h>
#include <sys/time.h>
#include "forms.h"

#include "matform.h"

/****
   This is a material editor.

   This is version 1.0.

   Written by Mark Overmars, Dept. of Computer Science, University of Utrecht,
   Email: markov@cs.ruu.nl
****/

typedef struct {
   Coord x,y,z;
} Point;

typedef struct {
    int numb;   /* number of vertices */
    Point v[4]; /* vertices */
    Point n[4]; /* normals */
} Patch;

typedef struct {
    int numb;     /* number of patches */
    Patch *pat;   /* the patches */
} OBJECT;


OBJECT glass;

/*************************
 Making Spin Objects
*************************/

#define PI    3.14159265
#define ZERO  0.00001

makespinobject(obj,smooth,rot,n,x1,z1,nx1,nz1,x2,z2,nx2,nz2)
/* makes a spin object */
OBJECT *obj;         /* the object to be created */
int smooth;          /* whether smooth or not */
int rot;             /* the number of rotation steps */
int n;               /* the number of segments to spin */
float x1[],z1[];     /* (x,z) coordinates of first endpoint of segments */
float nx1[],nz1[];   /* their normals */
float x2[],z2[];     /* (x,z) coordinates of second endpoint of segments */
float nx2[],nz2[];   /* their normals */
{
    float th,dth,a1,a2;
    Patch *p;
    int i,j;
    obj->numb = rot*n;
    p = obj->pat = (Patch *) malloc(obj->numb * sizeof(Patch));
    dth = 2.0*PI/rot;
    for (i=0; i<n; i++)
    {
        for (j=0; j<rot; j++)
        {
 	    th =  dth * j;   
	    if (smooth) {a1 = th; a2 =th+dth;} else { a1 = a2 = th +dth/2.0;};
	    if (x1[i] < ZERO || x2[i] <ZERO) {p->numb=3;} else {p->numb = 4;};
	    p->v[0].x = x1[i]*sin(th);     p->n[0].x = nx1[i]*sin(a1);
	    p->v[0].y = x1[i]*cos(th);     p->n[0].y = nx1[i]*cos(a1);
	    p->v[0].z = z1[i];             p->n[0].z = nz1[i];
	    p->v[1].x = x1[i]*sin(th+dth); p->n[1].x = nx1[i]*sin(a2);
	    p->v[1].y = x1[i]*cos(th+dth); p->n[1].y = nx1[i]*cos(a2);
	    p->v[1].z = z1[i];             p->n[1].z = nz1[i];
	    p->v[2].x = x2[i]*sin(th+dth); p->n[2].x = nx2[i]*sin(a2);
	    p->v[2].y = x2[i]*cos(th+dth); p->n[2].y = nx2[i]*cos(a2);
	    p->v[2].z = z2[i];             p->n[2].z = nz2[i];
	    p->v[3].x = x2[i]*sin(th);     p->n[3].x = nx2[i]*sin(a1);
	    p->v[3].y = x2[i]*cos(th);     p->n[3].y = nx2[i]*cos(a1);
	    p->v[3].z = z2[i];             p->n[3].z = nz2[i];
	    if (x1[i] <ZERO) 
	    {
		p->v[1] = p->v[2]; p->n[1] = p->n[2];
		p->v[2] = p->v[3]; p->n[2] = p->n[3];
	    };
	    p++;
	}
    }
}

makesmoothspinobject(obj,rot,n,x,z)
/* makes a smooth, connected spin object */
OBJECT *obj;         /* the object to be created */
int rot;             /* the number of rotation steps */
int n;               /* the number of points */
float x[],z[];       /* (x,z) coordinates of endpoints of segments */
{
    float x1[20],z1[20],nx1[20],nz1[20],x2[20],z2[20],nx2[20],nz2[20],l[20];
    int i;
    /* fill in the points */
    x1[0] = x[0]; z1[0] = z[0];
    for (i=1; i<n ; i++)
    	{ x1[i] = x2[i-1] = x[i]; z1[i] = z2[i-1] = z[i]; };
    /* fill in the normals */
    for (i=0; i<n-1 ; i++)
      {l[i] = sqrt((x[i+1]-x[i])*(x[i+1]-x[i]) + (z[i+1]-z[i])*(z[i+1]-z[i]));};
    nx1[0]   = 0.0; nz1[0]   =  1.0;
    nx2[n-2] = 0.0; nz2[n-2] = -1.0;
    for (i=1; i<n-1 ; i++)
    {
	nx1[i] = nx2[i-1] = ((z[i-1]-z[i])/l[i-1] + (z[i] - z[i+1])/l[i] )/2.0;
	nz1[i] = nz2[i-1] = ((x[i]-x[i-1])/l[i-1] + (x[i+1] - x[i])/l[i] )/2.0;
    }
    /* make it */
    makespinobject(obj,1,rot,n-1,x1,z1,nx1,nz1,x2,z2,nx2,nz2);
}

/*Making some nice objects */

float x1[20],z1[20];

makeglass(n)
int n;
{
    x1[ 0] = 0.000; z1[ 0] = 0.575;
    x1[ 1] = 0.050; z1[ 1] = 0.575;
    x1[ 2] = 0.100; z1[ 2] = 0.600;
    x1[ 3] = 0.200; z1[ 3] = 0.700;
    x1[ 4] = 0.250; z1[ 4] = 0.800;
    x1[ 5] = 0.250; z1[ 5] = 1.000;
    x1[ 6] = 0.275; z1[ 6] = 1.000;
    x1[ 7] = 0.275; z1[ 7] = 0.800;
    x1[ 8] = 0.250; z1[ 8] = 0.700;
    x1[ 9] = 0.200; z1[ 9] = 0.625;
    x1[10] = 0.100; z1[10] = 0.550;
    x1[11] = 0.050; z1[11] = 0.500;
    x1[12] = 0.025; z1[12] = 0.450;
    x1[13] = 0.025; z1[13] = 0.100;
    x1[14] = 0.050; z1[14] = 0.050;
    x1[15] = 0.250; z1[15] = 0.025;
    x1[16] = 0.250; z1[16] = 0.000;
    x1[17] = 0.225; z1[17] = 0.000;
    x1[18] = 0.000; z1[18] = 0.000;
    makesmoothspinobject(&glass,n,19,x1,z1);
}
    
drawobject(obj,wire,flat)
/* Draws a defined object. wire indicates whether wireframe. 
   flat indicates whether flat shaded. */
OBJECT obj;
int wire,flat;
{
    int i,j;
    Patch *p;
    p = obj.pat;
    if (flat) {shademodel(FLAT);} else {shademodel(GOURAUD);};
    for (i=0; i < obj.numb ; i++)
    {
	if (wire) {bgnclosedline();} else {bgnpolygon();};
	for (j=0; j < p->numb; j++)
	    { n3f(&p->n[j]); v3f(&p->v[j]);};
	if (wire) {endclosedline();} else {endpolygon();};
	p++;
    }
}

/* ------------------------------- */

long win;                /* the window identifiers */
float mater[21]; 	 /* the material */
float time_passed;

/* The lightsources. */
float light1[] = {LCOLOR,1.0,1.0,1.0,POSITION,10.0,10.0,5.0,0.0,LMNULL};
float light2[] = {LCOLOR,1.0,1.0,1.0,POSITION,-10.0,10.0,5.0,0.0,LMNULL};

/* The lightmodel. */
float model[] = {AMBIENT,0.2,0.2,0.2,LMNULL};

drawit()
/* draws the  object. */
{
    lmdef(DEFMATERIAL, 1, 21, mater);
    backface(1);
    zclear();
    RGBcolor(150,150,255);
    clear();
    pushmatrix();
    rot(time_passed*3.0,'X'); rot(time_passed*4.0,'Y');rot(time_passed*5.0,'Z');
    lmbind(MATERIAL,1); drawobject(glass,0,0);
    popmatrix();
}

/* The identity matrix. */
Matrix idmat = {1.0,0.0,0.0,0.0,
	        0.0,1.0,0.0,0.0,
		0.0,0.0,1.0,0.0,
		0.0,0.0,0.0,1.0};

initwindow()
/* Initializes all settings for a window. */
{
    makeglass(18);
    prefposition(850,1200,500,850);
    win = winopen("The View");
    zbuffer(TRUE);
    doublebuffer();
    RGBmode();
    blendfunction(BF_SA,BF_MSA);
    gconfig();
    mmode(MVIEWING);
    perspective(400,1.0,1.0,34.0);
    loadmatrix(idmat);
    lookat(0.0,0.0,4.0,0.0,0.0,0.0,0.0);
    /* define materials and lights */
    lmdef(DEFLIGHT, 1, 10, light1);
    lmdef(DEFLIGHT, 2, 10, light2);
    lmdef(DEFLMODEL, 1, 5, model);
    lmbind(LIGHT0,1);
    lmbind(LIGHT1,2);
    lmbind(LMODEL,1);
}

void print_cb(FL_OBJECT obj, long arg)
{
  printf("\nfloat %s[] = {\n",fl_show_input("Give material name:","Material"));
  printf("   AMBIENT, %.3f, %.3f, %.3f,\n", fl_get_slider_value(slar),
     fl_get_slider_value(slag), fl_get_slider_value(slab));
  printf("   DIFFUSE, %.3f, %.3f, %.3f,\n", fl_get_slider_value(sldr),
     fl_get_slider_value(sldg), fl_get_slider_value(sldb));
  printf("   EMISSION, %.3f, %.3f, %.3f,\n", fl_get_slider_value(sler),
     fl_get_slider_value(sleg), fl_get_slider_value(sleb));
  printf("   SPECULAR, %.3f, %.3f, %.3f,\n", fl_get_slider_value(slsr),
     fl_get_slider_value(slsg), fl_get_slider_value(slsb));
  printf("   SHININESS, %.3f,\n", fl_get_slider_value(slsh));
  printf("   ALPHA, %.3f };\n\n", fl_get_slider_value(slal));
}

mainloop()
{
   FL_OBJECT *obj;
   struct timeval tp;
   struct timezone tzp;
   long passed,lastsec,lasttime;
   float speed = 5.0 ;
   gettimeofday(&tp,&tzp);
   lastsec = tp.tv_sec;
   lasttime = tp.tv_usec;
   while(obj != exitobj)
   {
     obj = fl_check_forms();
     gettimeofday(&tp,&tzp);
     passed = 1000000 * (tp.tv_sec - lastsec) + (tp.tv_usec - lasttime);
     time_passed = time_passed + speed * passed/250000.0;
     lastsec = tp.tv_sec;
     lasttime = tp.tv_usec;
     mater[0] = AMBIENT;
     mater[1] = fl_get_slider_value(slar);
     mater[2] = fl_get_slider_value(slag);
     mater[3] = fl_get_slider_value(slab);
     mater[4] = DIFFUSE;
     mater[5] = fl_get_slider_value(sldr);
     mater[6] = fl_get_slider_value(sldg);
     mater[7] = fl_get_slider_value(sldb);
     mater[8] = SPECULAR;
     mater[9] = fl_get_slider_value(slsr);
     mater[10] = fl_get_slider_value(slsg);
     mater[11] = fl_get_slider_value(slsb);
     mater[12] = EMISSION;
     mater[13] = fl_get_slider_value(sler);
     mater[14] = fl_get_slider_value(sleg);
     mater[15] = fl_get_slider_value(sleb);
     mater[16] = SHININESS;
     mater[17] = fl_get_slider_value(slsh);
     mater[18] = ALPHA;
     mater[19] = fl_get_slider_value(slal);
     mater[20] = LMNULL;
     winset(win);
     reshapeviewport();
     drawit();
     swapbuffers();
   }
}


main()
{
   initwindow();
   create_the_forms();
   fl_set_slider_value(slar,0.2);
   fl_set_slider_value(slag,0.2);
   fl_set_slider_value(slab,0.2);
   fl_set_slider_value(sldr,0.8);
   fl_set_slider_value(sldg,0.8);
   fl_set_slider_value(sldb,0.8);
   fl_set_slider_value(slsr,0.0);
   fl_set_slider_value(slsg,0.0);
   fl_set_slider_value(slsb,0.0);
   fl_set_slider_value(sler,0.0);
   fl_set_slider_value(sleg,0.0);
   fl_set_slider_value(sleb,0.0);
   fl_set_slider_bounds(slsh,0.01,64.0);
   fl_set_slider_precision(slsh,0);
   fl_set_slider_value(slsh,5.0);
   fl_set_slider_value(slal,1.0);
   prefposition(50,800,500,850);
   fl_show_form(form,FL_PLACE_FREE,FALSE,NULL);
   mainloop();
   fl_hide_form(form);
}
